from .. import Application
from .. import __version__
from .. import handlers
from . import commands

app = Application(
    name="kapow",
    version="__version__",
    cli_handler=handlers.docopt_handler(commands.__doc__),
    pre_config_handler=None,
    config_handler=None,
    post_config_handler=None,
    pre_logging_config_handler=None,
    logging_config_handler=None,
    context_handler=None,
    command_handler=handlers.docopt_command_finder(commands),
)

main = app.main



import logging
import logging.config
from importlib.resources import read_text
from inspect import getmembers
from inspect import isfunction
from inspect import ismodule
from os import environ
from pathlib import Path
from pathlib import PosixPath
from pathlib import WindowsPath
from types import ModuleType
from types import SimpleNamespace
from typing import Any
from typing import Callable
from typing import Union
import tomlkit
from docopt import docopt
from tomlkit import comment
from tomlkit import document
from tomlkit import table
from . import confirm
from . import resources
from .appdirs import AppDirs
from .console import console


def cli_handler(app: "Application", ctx: Union[SimpleNamespace, Any]):
    ctx.cli_args = app.cli_args
    return app, ctx


def docopt_handler(docs):
    def _docopt_handler(app: "Application", ctx: Union[SimpleNamespace, Any]):
        ctx.cli_args = docopt(docs, app.cli_args, version=app.version)
        return app, ctx

    return _docopt_handler


def env_handler(app: "Application", ctx: Union[SimpleNamespace, Any]):
    env_name = f"{app.name.upper()}_"
    ctx.env_vars = {}
    for key, value in environ.items():
        if key.startswith(env_name):
            ctx.env_vars[key] = value
    return app, ctx


def appdir_handler(app: "Application", ctx: Union[SimpleNamespace, Any]):
    appdirs = AppDirs(app.name)
    ctx.dirs = app.context_class()
    ctx.dirs.app_home = appdirs.user_data_dir
    ctx.dirs.log_dir = appdirs.user_log_dir
    ctx.current_user = appdirs.user_name.lower()

    confirm.directory_exists(ctx.dirs.app_home)
    confirm.directory_exists(ctx.dirs.log_dir)

    ctx.files = app.context_class()
    ctx.files.config = Path(ctx.dirs.app_home, f"{app.name}.config.ini")
    ctx.files.logging_config = Path(ctx.dirs.app_home, f"{app.name}.logging.ini")
    ctx.files.log_file = Path(ctx.dirs.log_dir, f"{app.name}.logs.txt")

    return app, ctx


def default_config_builder(ctx):
    doc = document()
    doc.add(comment("default toml configuration file"))
    app = table()
    app["debug"] = True
    app["wrk_dir"] = str(ctx.dirs.app_home / "wrk_dir")
    doc["app"] = app
    ctx.files.config.write_text(tomlkit.dumps(doc))


def pre_config_handler(config_builder):
    def _pre_config_handler(app: "Application", ctx: Union[SimpleNamespace, Any]):
        confirm.ctx_var(ctx, "files.config", (Path, PosixPath, WindowsPath))

        if not ctx.files.config.exists():
            config_builder(ctx)
        return app, ctx

    return _pre_config_handler


def config_handler(app: "Application", ctx: Union[SimpleNamespace, Any]):
    """
    Reads toml config file format.

    :param app:
    :param ctx:
    :return:
    """
    confirm.expr(
        ctx.files.config.exists(), f"Config file does not exist: {ctx.files.config}"
    )
    ctx.config = tomlkit.loads(ctx.files.config.read_text())
    return app, ctx


def post_config_handler(config_validator):
    """
    Using the provided config_validator, validate the config data.

    config_validator should raise a LaunchError indicating the issue with the config data.

    :param config_validator:
    :return: Application, Context

    """

    def _post_config_handler(app: "Application", ctx: Union[SimpleNamespace, Any]):
        config_validator(ctx.config)
        return app, ctx

    return _post_config_handler


def context_handler(app: "Application", ctx: Union[SimpleNamespace, Any]):
    """
    The purpose of the context handler is to coerce the context object into
    a state for the application. This might involve merging cli and config
    arguments into single values, adding new variables based on those inputs,
    removing temporary values from the context object or re-writing to a new
    context object.

    :param app: Application
    :param ctx: Context
    :return: Application, Context
    """
    return app, ctx


def default_logging_config_builder(app, ctx):
    log_cfg_txt = read_text(resources, "logging.ini")
    ctx.files.logging_config.write_text(
        log_cfg_txt.format(logfile=ctx.files.log_file, appname=app.name)
    )


def pre_logging_config_handler(logging_config_builder):
    def _pre_logging_config_handler(
        app: "Application", ctx: Union[SimpleNamespace, Any]
    ):
        confirm.ctx_var(ctx, "files.logging_config", (Path, PosixPath, WindowsPath))

        if not ctx.files.logging_config.exists():
            logging_config_builder(app, ctx)

        if not ctx.files.log_file.exists():
            ctx.files.log_file.write_text("")

        return app, ctx

    return _pre_logging_config_handler


def logging_config_handler(app: "Application", ctx: Union[SimpleNamespace, Any]):
    # TODO: we need an option for pointing to an alternative logging config file
    confirm.ctx_var(ctx, "files.logging_config", (Path, PosixPath, WindowsPath))

    logging.config.fileConfig(ctx.files.logging_config)
    app.log = logging.getLogger(app.name)

    return app, ctx


def command_handler(command_finder: Callable) -> Callable:
    def _command_handler(app: "Application", ctx: Union[SimpleNamespace, Any]):
        app.command = command_finder(ctx)
        return app, ctx

    return _command_handler


def error_handler(app: "Application", ctx: Union[SimpleNamespace, Any], error):
    """
    This is a special case handler that is called as the
    top level exception handler. It is called from within the `execute_handler`.

    In addition to the application and context arguments, it also takes an error argument.

    :param app: Application object
    :param ctx:  Context object
    :param error: Exception object
    :return: None

    """
    import traceback
    from rich import box
    from rich.panel import Panel

    console.print(
        f"\n  [green]{app.name}[/green] failed with error: [red]{error}[/red]\n"
    )
    console.print(Panel(traceback.format_exc().strip(), box.SQUARE, highlight=True))


# TODO: rename to main_handler?
def execute_handler(app: "Application") -> Callable:
    """
    This is a special case handler that is the final function called
    in the kapower pipeline. It is responsible for creating
    the application's main function, which runs the application.

    It only takes an application reference, and is responsible for
    creating the context object, and for handling the top-level
    errors.

    In practice user's should not be overriding the execute handler.

    :param app: Application
    :return:
    """

    def _main():
        nonlocal app
        context = app.context_class()
        for handler_key in app._execution_order:
            try:
                handler = app._handlers[handler_key]
                app, context = handler(app, context)
            except Exception as ex:
                app.error_handler(app, context, ex)
                return

        try:
            app.command(context)
        except Exception as ex:
            app.error_handler(app, context, ex)

    return _main


def docopt_command_finder(cmd_obj: Union[ModuleType, Callable, SimpleNamespace]):
    def match_func_name_to_cli_cmd(func_name, cli_args):
        possible_names = [func_name]
        possible_names.append(func_name.replace("_", "."))
        possible_names.append(func_name.replace("_", "-"))
        for name in possible_names:
            if name in cli_args and cli_args[name] is True:
                return True
        return False

    def _docopt_command_finder(app, ctx):
        confirm.ctx_var(ctx, "cli_args", dict)

        if isfunction(cmd_obj) or callable(cmd_obj):
            app.command = cmd_obj

        elif ismodule(cmd_obj) or isinstance(cmd_obj, SimpleNamespace):
            functions = [f for f in getmembers(cmd_obj) if isfunction(f[1])]
            for func_name, func_obj in functions:
                if match_func_name_to_cli_cmd(func_name, ctx.cli_args):
                    app.command = func_obj
                    break

        return app, ctx

    return _docopt_command_finder



import inspect
import sys
from pathlib import Path
from typing import Callable
from .errors import LaunchError



def expr(expression_result: bool, assertion_message: str):
    if expression_result is False:
        raise LaunchError(assertion_message)


def ctx_var(ctx, path: str, var_type):
    var_path = []
    root = ctx
    for var in path.split("."):
        var_path.append(var)
        root = getattr(root, var, None)
        if not root:
            raise LaunchError(
                f"Context missing expected value at: {'.'.join(var_path)}."
            )
    if not isinstance(root, var_type):
        raise LaunchError(
            f"Expecting context variable {'.'.join(var_path)} to be type {var_type}. Found {type(root)}."
        )


def directory_exists(dirpath: Path):
    """
    Create the directories in the path if they do not exist.

    :param dirpath:
    """
    if dirpath.is_file():
        dirpath = dirpath.parent
    dirpath.mkdir(parents=True, exist_ok=True)


def _assert_callable(handler: Callable):
    if not callable(handler):
        raise LaunchError(f"Provided object is not callable: {handler}.")


def _assert_signature(handler: Callable, expect: int = 2):
    _assert_callable(handler)
    signature: inspect.Signature = inspect.signature(handler)
    param_names = signature.parameters.keys()
    param_cnt = len(param_names)
    if param_cnt != expect:
        argument = "argument" if param_cnt == 1 else "arguments"
        raise LaunchError(
            f"A Launch handler callable must have 2 arguments (app, ctx). `{handler.__name__}` has {param_cnt} {argument}."
        )


def handler_func(handler: Callable):
    _assert_signature(handler, expect=2)


def error_func(handler: Callable):
    _assert_signature(handler, expect=3)


def command_func(handler: Callable):
    _assert_signature(handler, expect=1)



"""
Usage:
  kapow init PATH

Options:
  -h --help      Show this help message.
  -v --version   Show app version.

"""
from pathlib import Path
from types import SimpleNamespace
from rich.prompt import Prompt, Confirm
from ..console import console
from ..appdirs import user_full_name
from . import project


def init(ctx):
    ns = SimpleNamespace()

    ns.project_dir = Path(ctx.cli_args['PATH']).resolve()

    pname = ns.project_dir.stem

    ns.project_name = Prompt.ask("Project name", default=pname)
    ns.author_name = Prompt.ask("Author name", default=user_full_name())
    ns.author_email = Prompt.ask("Author email", default="")
    ns.project_type = Prompt.ask("Packaging", choices=['setup', 'poetry'], default='poetry')
    ns.src_dir = Confirm.ask("Use src dir?", default=False)


    console.print("\nProject details:")
    console.print(f"  name: [green]{ns.project_name}[/green]")
    console.print(f"  directory: [green]{ns.project_dir}[/green]")
    console.print(f"  project type: [green]{ns.project_type}[/green]")
    if ns.src_dir:
        console.print("  use src dir: [green]yes[/green]")
    else:
        console.print("  use src dir: [red]no[/red]")
    console.print()

    if Confirm.ask("Generate project"):
        project.generate(ns)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['kapow', 'kapow.cli', 'kapow.cli.resources', 'kapow.resources']

package_data = \
{'': ['*']}

install_requires = \
['docopt-ng>=0.7.2,<0.8.0', 'rich>=11.2.0,<12.0.0', 'tomlkit>=0.9.2,<0.10.0']

entry_points = \
{'console_scripts': ['kapow = kapow.cli:main']}

setup_kwargs = {
    'name': 'kapow',
    'version': '0.1.0',
    'description': '',
    'long_description': '### kapow\n\n\nA python application launch framework with a punch! Kapow!\n\n**This alpha software currently under active development**\n',
    'author': 'Mark Gemmill',
    'author_email': 'mark@markgemmill.com',
    'maintainer': 'Mark Gemmill',
    'maintainer_email': 'dev@markgemmill.com',
    'url': 'https://github.com/markgemmill/kapow',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from pathlib import Path
import shutil
from tempfile import TemporaryDirectory
from typing import Protocol, List
from importlib.resources import read_text
from kapow.console import console
from . import resources
import tomlkit


class GenerateContext(Protocol):
    project_dir: Path


def file_from_template(ctx, filepath, resource_name):
    template = read_text(resources, resource_name)
    filepath.write_text(template.format(
        appname=ctx.project_name,
    ))


def create_poetry_pyproject(ctx: GenerateContext, pyproject_file: Path):
    doc = tomlkit.document()

    poetry = tomlkit.table()
    poetry["name"] = ctx.project_name
    poetry["version"] = "0.1.0"
    poetry["description"] = ""

    author = ""
    if ctx.author_name:
        author = ctx.author_name
    if ctx.author_email:
        author += f" <{ctx.author_email}>"

    poetry["authors"] = [author.strip()]

    deps = tomlkit.table()
    deps["python"] = "^3.9"
    deps["tomlkit"] = "^0.9.2"
    deps["docopt-ng"] = "^0.7.2"
    deps["rich"] = "^11.2"

    build_system = tomlkit.table()
    build_system["requires"] = ["poetry-core>=1.0.0"]
    build_system["build-backend"] = "poetry.core.masonry.api"

    scripts = tomlkit.table()
    scripts[ctx.project_name] = f"{ctx.project_name}.cli:main"

    doc["tool"] = tool = tomlkit.table()
    tool["poetry"] = poetry
    tool["poetry"]["dependencies"] = deps
    tool["poetry"]["scripts"] = scripts
    doc["build-system"] = build_system

    pyproject_file.write_text(tomlkit.dumps(doc))


def generate(ctx: GenerateContext):

    if ctx.project_dir.exists() and [f for f in ctx.project_dir.iterdir()]:
        console.print("\n[red]The project directory is not empty.[/red]")
        console.print("Select a new or empty directory when creating a project.\n")
        return

    with TemporaryDirectory() as root_dir:

        src_dir = Path(root_dir)

        if ctx.src_dir:
            src_dir = Path(root_dir, 'src')

        src_dir = Path(src_dir, ctx.project_name)
        src_dir.mkdir(parents=True, exist_ok=True)

        # define project files
        readme_file = Path(root_dir, 'README.md')
        pyproject_file = Path(root_dir, 'pyproject.toml')
        init_py = Path(src_dir, '__init__.py')
        cli_file = Path(src_dir, 'cli.py')
        commands_file = Path(src_dir, 'commands.py')

        # generate file content
        create_poetry_pyproject(ctx, pyproject_file)
        file_from_template(ctx, init_py, 'init.txt')
        file_from_template(ctx, cli_file, 'cli.txt')
        file_from_template(ctx, commands_file, 'commands.txt')

        readme_file.touch()

        if not ctx.project_dir.exists():
            ctx.project_dir.mkdir(parents=True, exist_ok=True)

        shutil.copytree(root_dir, ctx.project_dir, dirs_exist_ok=True)



import inspect
import sys
from pathlib import Path
from types import SimpleNamespace
from typing import Callable
from typing import ClassVar
from typing import Union
from . import confirm
from . import handlers
from .errors import LaunchError

__version__ = "0.1.0"


class Application:
    """
    The Application object in essence is a specialized pipline manager.

    Every application requires a series of resources in order to run, many of which
    are common across all applications.

    The most simple scripts do not need anything other than a main function, but beyond that
    programs will typically need to pull in argument and options from the command line, configuration files,
    environment variable; define their execution environment and access to local resource; will want to configure logging;
    and finally select the appropriate entry function.

    Launch defines an ordered series of events:

    cli: parse and convert command line arguments
    environ variable: read in environment variables
    configuration: read configuration files or resource
    merge arguments: merge or override cli/environ/config arguments
    configure logging: set logging configurations
    select application's main function based on program cli/env/config
    define final error handling

    """

    def __init__(
        self,
        name: str,
        version: str,
        context_class: ClassVar = SimpleNamespace,
        cli_handler: Union[bool, Callable, None] = True,
        env_handler: Union[bool, Callable, None] = True,
        appdir_handler: Union[bool, Callable, None] = True,
        pre_config_handler: Union[Callable, None] = None,
        config_handler: Union[bool, Callable, None] = True,
        post_config_handler: Union[Callable, None] = None,
        context_handler: Union[bool, Callable, None] = True,
        pre_logging_config_handler: Union[bool, Callable, None] = True,
        logging_config_handler: Union[bool, Callable, None] = True,
        command_handler: Union[bool, Callable, None] = True,
        error_handler: Union[Callable, None] = None,
        execute_handler: Union[Callable, None] = None,
        **kwargs,
    ):
        self.name = name
        self.version = version
        self.context_class = context_class
        # these handlers are mandatory
        self.execute_handler = handlers.execute_handler
        self.error_handler = handlers.error_handler

        self._handlers = {}
        self._execution_order = []

        self._add_handler("cli_handler", cli_handler, handlers.cli_handler)
        self._add_handler("env_handler", env_handler, handlers.env_handler)
        self._add_handler("appdir_handler", appdir_handler, handlers.appdir_handler)
        self._add_handler("pre_config_handler", pre_config_handler)
        self._add_handler("config_handler", config_handler, handlers.config_handler)
        self._add_handler("post_config_handler", post_config_handler)
        self._add_handler("context_handler", context_handler, handlers.context_handler)
        self._add_handler(
            "pre_logging_config_handler",
            pre_logging_config_handler,
            handlers.pre_logging_config_handler(
                handlers.default_logging_config_builder
            ),
        )
        self._add_handler(
            "logging_config_handler",
            logging_config_handler,
            handlers.logging_config_handler,
        )
        self._add_handler("command_handler", command_handler, handlers.command_handler)

        # special case
        self._add_handler("error_handler", error_handler, handlers.error_handler)
        self._add_handler("execute_handler", execute_handler, handlers.execute_handler)

        for name, handler in kwargs.items():
            if name.startswith("before_") or name.startswith("after_"):
                self._add_handler(name, handler)

        # for testing purposes
        if "cli_args" in kwargs:
            self._cli_args = kwargs["cli_args"]

    def _add_handler(self, name, handler, default=None):

        # do not load the handler
        if handler is None:
            return

        # load the default handler
        if handler is True:
            if not default:
                raise LaunchError("Using True requires a default handler.")
            self._handlers[name] = default
            self._execution_order.append(name)
            return

        if name == "error_handler" and handler is not None:
            confirm.error_func(handler)
            self.error_handler = handler
            return

        if name == "execute_handler" and handler is not None:
            confirm.command_func(handler)
            self.execute_handler = handler
            return

        confirm.handler_func(handler)

        # insert a user defined handler - before or after the named standard handler
        if name.startswith("before_") or name.startswith("after_"):

            index = 0

            if name.startswith("before_"):
                relative_to_name = name.replace("before_", "")

            elif name.startswith("after_"):
                relative_to_name = name.replace("after_", "")
                index = 1

            if (
                relative_to_name not in self._handlers
                or relative_to_name not in self._execution_order
            ):
                raise LaunchError(
                    f"Expecting to insert `{name}` relative to `{relative_to_name}`, but `{relative_to_name}` does not exist."
                )

            insert_at = self._execution_order.index(relative_to_name)

            self._handlers[name] = handler
            self._execution_order.insert(insert_at + index, name)
            return

        # load the default handler
        self._handlers[name] = handler
        self._execution_order.append(name)

    @property
    def cli_args(self):
        if hasattr(self, '_cli_args'):
            return self._cli_args
        return sys.argv[1:]

    @property
    def main(self) -> Callable:
        """

        :return: main function
        """
        return self.execute_handler(self)
